"""
Module for Terraform Cloud API Endpoint: Org Tags.
"""

from .endpoint import TFCEndpoint

class TFCOrgTags(TFCEndpoint):
    """
    `Org Tags API Docs \
        <https://www.terraform.io/docs/cloud/api/organization-tags.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._org_tags_base_url = \
            f"{self._api_v2_base_url}/organizations/{org_name}/tags"
        self._tags_base_url = \
            f"{self._api_v2_base_url}/tags"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def list_tags(self):
        """
        ``GET /organizations/:organization_name/tags``

        `Get Tags API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/organization-tags.html#get-tags>`_
        """
        # TODO: does this support pagination?
        return self._get(self._org_tags_base_url)

    def delete_tags(self, payload):
        """
        ``DELETE /organizations/:organization_name/tags``

        `Delete Tags API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/organization-tags.html#delete-tags>`_

        `Delete Tags API Doc Sample Payload \
            <https://www.terraform.io/docs/cloud/api/organization-tags.html#sample-payload>`_
        """
        return self._delete(self._org_tags_base_url, data=payload)

    def add_workspaces_to_tag(self, tag_id, payload):
        """
        ``POST /tags/:tag_id/relationships/workspaces``

        `Add Workspace to a Tag API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/organization-tags.html#add-workspaces-to-a-tag>`_

        `Add Workspace to a Tag API Doc Sample Payload \
            <https://www.terraform.io/docs/cloud/api/organization-tags.html#sample-payload-1>`_
        """
        url = f"{self._tags_base_url}/{tag_id}/relationships/workspaces"
        return self._post(url, data=payload)

    # NOTE: this endpoint has been temporarily removed
    """
    def remove_workspaces_from_tag(self, tag_id, payload):
        \"""
        ``DELETE /tags/:tag_id/relationships/workspaces``

        `Remove Workspaces from Tags API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/organization-tags.html#remove-workspaces-from-a-tag>`_

        `Remove Workspaces from Tags API Doc Sample Payload \
            <>`_
        \"""
        url = f"{self._tags_base_url}/{tag_id}/relationships/workspaces"
        return self._delete(url, data=payload)
    """



"""
Module for testing the Terraform Cloud API Endpoint: Workspaces.
"""

from .base import TestTFCBaseTestCase
from ._constants import PAGE_START, PAGE_SIZE


class TestTFCWorkspaces(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Workspaces.
    """

    _unittest_name = "ws"
    _endpoint_being_tested = "workspaces"

    def setUp(self):
        # Add an SSH Key to TFC
        created_key = self._api.ssh_keys.create(self._get_ssh_key_create_payload())["data"]
        self._created_key_id = created_key["id"]

    def tearDown(self):
        self._api.ssh_keys.destroy(self._created_key_id)

    def test_workspaces(self):
        """
        Test the Workspaces API endpoints.
        """

        # Get the number of existing workspaces, then create one and compare them
        workspace = self._api.workspaces.create(
            self._get_ws_no_vcs_create_payload())["data"]
        ws_id = workspace["id"]

        # TODO: use constants for the page sizes and number?
        # List the workspaces, confirm we have the included values
        listed_ws_raw = self._api.workspaces.list(page=PAGE_START, page_size=PAGE_SIZE, include=["organization"])
        self.assertIn("included", listed_ws_raw)

        listed_ws = listed_ws_raw["data"]
        found_ws = False
        for workspace in listed_ws:
            if workspace["id"] == ws_id:
                found_ws = True
                break
        self.assertTrue(found_ws)

        # List all the workspaces, confirm we have the included values
        all_ws = self._api.workspaces.list_all(include=["organization"])
        self.assertIn("included", all_ws)

        found_ws = False
        for workspace in all_ws["data"]:
            if workspace["id"] == ws_id:
                found_ws = True
                break
        self.assertTrue(found_ws)

        # Lock the workspace and confirm it's locked
        ws_locked = self._api.workspaces.lock(
            ws_id, {"reason": "Unit testing."})["data"]
        self.assertTrue(ws_locked["attributes"]["locked"])

        # Unlock the workspace and confirm it's unlocked
        ws_unlocked = self._api.workspaces.unlock(ws_id)
        self.assertFalse(ws_unlocked["data"]["attributes"]["locked"])

        # Relock the workspace and confirm it's locked
        ws_relocked = self._api.workspaces.lock(
            ws_id, {"reason": "Unit testing."})["data"]
        self.assertTrue(ws_relocked["attributes"]["locked"])

        # Force an unlock and confirm its unlocked
        ws_forced = self._api.workspaces.force_unlock(ws_id)["data"]
        self.assertFalse(ws_forced["attributes"]["locked"])

        # Update the workspace, check that the updates took effect
        updated_name = self._unittest_random_name()
        update_payload = {
            "data": {
                "type": "workspaces",
                "attributes": {
                    "name": updated_name
                }
            }
        }
        ws_updated = self._api.workspaces.update(update_payload, workspace_id=ws_id)["data"]
        self.assertEqual(
            updated_name, ws_updated["attributes"]["name"])

        # Assign an SSH key and confirm it's added
        assign_payload = {
            "data": {
                "attributes": {
                    "id": self._created_key_id
                },
                "type": "workspaces"
            }
        }
        self._api.workspaces.assign_ssh_key(ws_id, assign_payload)

        # Show by ID, and make sure we get the right payload back. Also check for the
        # the newly assigned SSH key
        ws_shown_by_id_raw = self._api.workspaces.show(workspace_id=ws_id, include=["organization"])
        self.assertIn("included", ws_shown_by_id_raw)
        ws_shown_by_id = ws_shown_by_id_raw["data"]
        self.assertEqual(ws_id, ws_shown_by_id["id"])
        self.assertIn("ssh-key", ws_shown_by_id["relationships"])

        # Unassign the SSH key and confirm it's removed
        unassign_payload = {
            "data": {
                "attributes": {
                    "id": None
                },
                "type": "workspaces"
            }
        }
        self._api.workspaces.unassign_ssh_key(ws_id, unassign_payload)
        # TODO: include test when include works?
        ws_shown_by_id = self._api.workspaces.show(workspace_id=ws_id)["data"]
        self.assertNotIn("ssh-key", ws_shown_by_id["relationships"])

        # Add tags to the workspace
        ws_add_tags_payload = {
            "data": [
                {
                    "type": "tags",
                    "attributes": {
                        "name": "foo"
                    }
                },
                {
                    "type": "tags",
                    "attributes": {
                        "name": "bar"
                    }
                }
            ]
        }
        self._api.workspaces.add_tags(ws_id, ws_add_tags_payload)

        # Get the tags and confirm that two were added to the workspace
        added_ws_tags = self._api.workspaces.list_tags(ws_id)["data"]
        self.assertEqual(len(added_ws_tags), len(ws_add_tags_payload["data"]))

        # Also test the list all tags function
        all_tags = self._api.workspaces.list_all_tags(ws_id)["data"]
        self.assertEqual(len(all_tags), len(ws_add_tags_payload["data"]))

        # Remove one tag from the workspace
        ws_remove_tags_payload = {
            "data": [
                {
                    "type": "tags",
                    "id": added_ws_tags[0]["id"]
                }
            ]

        }
        self._api.workspaces.remove_tags(ws_id, ws_remove_tags_payload)

        # Get the tags and confirm the only tag left is the one we didn't remove.
        current_ws_tags = self._api.workspaces.list_tags(ws_id)["data"]
        self.assertEqual(current_ws_tags[0]["id"], added_ws_tags[-1]["id"])

        self._api.workspaces.destroy(workspace_name=updated_name)
        listed_ws_raw = self._api.workspaces.list(include=["organization"])
        listed_ws = listed_ws_raw["data"]
        found_ws = False
        for workspace in listed_ws:
            if workspace["id"] == ws_id:
                found_ws = True
                break

        self.assertFalse(found_ws)

    def test_workspaces_remote_state_consumers(self):
        """
        Test the Workspaces Remote State Consumers API endpoints.
        """

        # Create 3 workspaces, one to add a consumer to, one to update with.
        workspace1_id = self._api.workspaces.create(self._get_ws_no_vcs_create_payload())["data"]["id"]
        workspace2_id = self._api.workspaces.create(self._get_ws_no_vcs_create_payload())["data"]["id"]
        workspace3_id = self._api.workspaces.create(self._get_ws_no_vcs_create_payload())["data"]["id"]

        gotten_consumers = self._api.workspaces.get_remote_state_consumers(workspace1_id)["data"]
        self.assertEqual(len(gotten_consumers), 0)

        # Add the second workspace as a consumer to the first workspace
        add_payload = {
            "data": [
                {
                    "id": workspace2_id,
                    "type": "workspaces"
                }
            ]
        }
        self._api.workspaces.add_remote_state_consumers(workspace1_id, add_payload)

        # Confirm the consumers
        gotten_consumers = self._api.workspaces.get_remote_state_consumers(workspace1_id)["data"]
        self.assertIn(workspace2_id, gotten_consumers[0]["relationships"]["remote-state-consumers"]["links"]["related"])

        # Update to add the third workspace as a consumer of the first workspace
        replace_delete_payload = {
            "data": [
                {
                    "id": workspace3_id,
                    "type": "workspaces"
                }
            ]
        }
        self._api.workspaces.replace_remote_state_consumers(workspace1_id, replace_delete_payload)

        # Confirm the consumers
        gotten_consumers = self._api.workspaces.get_remote_state_consumers(workspace1_id)["data"]
        self.assertIn(workspace3_id, gotten_consumers[0]["relationships"]["remote-state-consumers"]["links"]["related"])

        self._api.workspaces.delete_remote_state_consumers(workspace1_id, replace_delete_payload)

        # Confirm the consumers
        gotten_consumers = self._api.workspaces.get_remote_state_consumers(workspace1_id)["data"]
        self.assertEqual(len(gotten_consumers), 0)

        # Destroy all workspaces
        self._api.workspaces.destroy(workspace_id=workspace1_id)
        self._api.workspaces.destroy(workspace_id=workspace2_id)
        self._api.workspaces.destroy(workspace_id=workspace3_id)



"""
Module for Terraform Cloud API Endpoint: Run Tasks.
"""

from .endpoint import TFCEndpoint

class TFCRunTasks(TFCEndpoint):
    """
    `Run Tasks API Docs \
        <https://www.terraform.io/docs/cloud/api/run-tasks.html>`_
    """
    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._org_event_hooks_base_url = f"{self._api_v2_base_url}/organizations/{self._org_name}/event-hooks"
        self._event_hooks_base_url = f"{self._api_v2_base_url}/event-hooks"
        self._tasks_base_url = f"{self._api_v2_base_url}/tasks"
        self._ws_api_v2_base_url = f"{self._api_v2_base_url}/workspaces"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        # TODO: remove this once it goes to TFE GA
        return True

    def terraform_enterprise_only(self):
        return False

    def create_event_hook(self, payload):
        """
        ``POST /organizations/:organization_name/event-hooks``

        `Run Tasks Create Event Hook API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run-tasks.html#create-an-event-hook>`_

        `Run Tasks Create Event Hook Sample Payload \
            <https://www.terraform.io/docs/cloud/api/run-tasks.html#sample-payload>`_
        """
        return self._create(self._org_event_hooks_base_url, payload)


    def list_event_hooks(self, page=None, page_size=None, include=None):
        """
        ``GET /organizations/:organization_name/event-hooks``

        `Run Tasks List Event Hooks API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run-tasks.html#list-event-hooks>`_

        Query Parameter(s) (`details \
            <https://www.terraform.io/docs/cloud/api/run-tasks.html#query-parameters>`__):
            - ``page_size`` (Optional)
            - ``page`` (Optional)
            - ``include`` (Optional)
        """
        return self._list(self._org_event_hooks_base_url, page=page, page_size=page_size, include=include)

    def list_all_event_hooks(self):
        """
        This function does not correlate to an endpoint in the TFC API Docs specifically,
        but rather is a helper function to wrap the `list` endpoint, which enumerates out
        every page so users do not have to implement the paging logic every time they just
        want to list every run trigger for a workspace.

        Returns an object with two arrays of objects.
        """
        return self._list_all(self._org_event_hooks_base_url)

    def show_event_hook(self, event_hook_id):
        """
        ``GET /event-hooks/:id``

        `Run Tasks Show Event Hook API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run-tasks.html#show-an-event-hook>`_
        """
        url = f"{self._event_hooks_base_url}/{event_hook_id}"
        return self._show(url)

    def update_event_hook(self, event_hook_id, payload):
        """
        ``PATCH /event-hooks/:id``

        `Run Tasks Update Event Hook API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run-tasks.html#update-a-task>`_

        `Update Event Hook Sample Payload \
            <https://www.terraform.io/docs/cloud/api/run-tasks.html#sample-payload-3>`_
        """
        url = f"{self._event_hooks_base_url}/{event_hook_id}"
        return self._update(url, payload)

    def destroy_event_hook(self, event_hook_id):
        """
        ``DELETE /event-hooks/:id``

        `Run Tasks Destroy Event Hook API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run-tasks.html#delete-an-event-hook>`_
        """
        url = f"{self._event_hooks_base_url}/{event_hook_id}"
        return self._destroy(url)

    def attach_event_hook_as_task(self, workspace_id, payload):
        """
        ``POST /workspaces/:workspace_id/tasks``

        `Run Tasks Attach Event Hook API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run-tasks.html#attach-an-event-hook-to-a-workspace-as-a-task>`_

        `Attach Event Hook Sample Payload \
            <https://www.terraform.io/docs/cloud/api/run-tasks.html#sample-payload-2>`_
        """
        url = f"{self._ws_api_v2_base_url}/{workspace_id}/tasks"
        return self._post(url, data=payload)

    def list(self, workspace_id, page=None, page_size=None):
        """
        ``GET /workspaces/:workspace_id/tasks``

        `Run Tasks List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run-tasks.html#list-tasks>`_

        Query Parameter(s) (`details \
            <https://www.terraform.io/docs/cloud/api/run-tasks.html#query-parameters-1>`__):
            - ``page_size`` (Optional)
            - ``page`` (Optional)
        """
        url = f"{self._ws_api_v2_base_url}/{workspace_id}/tasks"
        return self._list(url, page=page, page_size=page_size)

    def show(self, task_id):
        """
        ``GET /tasks/:id``

        `Run Tasks Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run-tasks.html#show-a-task>`_
        """
        url = f"{self._tasks_base_url}/{task_id}"
        return self._show(url)

    def update(self, task_id, payload):
        """
        ``PATCH /tasks/:id``

        `Run Tasks Update API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run-tasks.html#update-a-task>`_

        `Run Tasks Update Sample Payload \
            <https://www.terraform.io/docs/cloud/api/run-tasks.html#sample-response-7>`_
        """
        url = f"{self._tasks_base_url}/{task_id}"
        return self._update(url, payload)

    def destroy(self, task_id):
        """
        ``DELETE /tasks/:id``

        `Run Tasks Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run-tasks.html#delete-a-task>`_
        """
        url = f"{self._tasks_base_url}/{task_id}"
        return self._destroy(url)



"""
Module for Terraform Cloud API Endpoint: VCS Events.
"""

from .endpoint import TFCEndpoint

class TFCVCSEvents(TFCEndpoint):
    """
    `VCS Events API Docs \
        <https://www.terraform.io/docs/cloud/api/vcs-events.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._vcs_events_api_v2_base_url = \
            f"{self._api_v2_base_url}/organizations/{org_name}/vcs-events"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        # TODO: change this once it is released in TFE.
        return True

    def terraform_enterprise_only(self):
        return False

    def list(self, page=None, page_size=None, include=None, filters=None):
        """
        ``GET /organizations/:organization_name/vcs-events``

        `VCS Events List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/vcs-events.html#list-vcs-events>`_

        Query Parameter(s) (`details \
            <https://www.terraform.io/docs/cloud/api/vcs-events.html#query-parameters>`__):
            - ``page[number]`` (Optional)
            - ``page[size]`` (Optional)
            - ``filter[from]`` (Optional)
            - ``filter[to]`` (Optional)
            - ``filter[oauth_client_external_ids]`` (Optional)
            - ``filter[levels]`` (Optional)
            - ``include`` (Optional)

        Example filter(s):

        .. code-block:: python

            filters = [
                {
                    "keys": ["from"],
                    "value": "foo"
                },
                {
                    "keys": ["to"],
                    "value": "bar"
                }
            ]

        NOTE: Currently, this feature is in beta, and only supports GitLab.
        Since all of the testing is done with GitHub, this function is
        provided, but is not well tested yet.
        """

        return self._list(self._vcs_events_api_v2_base_url, \
            page=page, page_size=page_size, include=include, filters=filters)



"""
Module for Terraform Cloud API Endpoint: Admin Orgs.
"""

from .endpoint import TFCEndpoint

class TFCAdminOrgs(TFCEndpoint):
    """
    `Admin Orgs API Docs \
        <https://www.terraform.io/docs/cloud/api/admin/organizations.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._org_api_v2_base_url = f"{self._api_v2_base_url}/admin/organizations"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return True

    def destroy(self, org_name):
        """
        ``DELETE /admin/organizations/:name``

        `Admin Destroy Org API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/organizations.html#delete-an-organization>`_
        """
        url = f"{self._org_api_v2_base_url}/{org_name}"
        return self._destroy(url)

    def list(self, include=None):
        """
        ``GET /api/v2/admin/organizations``

        `Admin List Orgs API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/organizations.html#list-all-organizations>`_

        This endpoint lists all organizations in the Terraform Cloud installation.
        """
        return self._list(self._org_api_v2_base_url, include=include)

    def list_org_module_consumers(self, org_name, page=None, page_size=None):
        """
        ``GET /api/v2/admin/organizations/:name/relationships/module-consumers``

        `Admin List Module Consumers for an Organization API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/organizations.html#list-module-consumers-for-an-organization>`_

        Query Parameter(s) (`details \
            <https://www.terraform.io/docs/cloud/api/admin/organizations.html#query-parameters-1>`__):
            - ``page`` (Optional)
            - ``page_size`` (Optional)
        """
        url = f"{self._org_api_v2_base_url}/{org_name}/relationships/module-consumers"
        return self._list(url, page=page, page_size=page_size)

    def show(self, org_name, include=None):
        """
        ``GET /api/v2/admin/organizations/:name``

        `Admin Show Org API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/organizations.html#show-an-organization>`_
        """
        url = f"{self._org_api_v2_base_url}/{org_name}"
        return self._show(url, include=include)

    def update(self, org_name, payload):
        """
        ``PATCH /admin/organizations/:name``

        `Admin Orgs Update API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/organizations.html#update-an-organization>`_

        `Update Sample Payload \
            <https://www.terraform.io/docs/cloud/api/admin/organizations.html#sample-payload>`_
        """
        url = f"{self._org_api_v2_base_url}/{org_name}"
        return self._update(url, payload)

    def update_org_module_consumers(self, org_name, payload):
        """
        ``PATCH /admin/organizations/:name/relationships/module-consumers``

        `Admin Orgs Update Module Consumers API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/organizations.html#update-an-organization-39-s-module-consumers>`_

        `Update Org Module Consumers Sample Payload \
            <https://www.terraform.io/docs/cloud/api/admin/organizations.html#sample-payload-1>`_
        """

        url = f"{self._org_api_v2_base_url}/{org_name}/relationships/module-consumers"

        return self._update(url, payload)



"""
Module for testing the Terraform Cloud API Endpoint: VCS Events.
"""

from .base import TestTFCBaseTestCase


class TestTFCVCSEvents(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: VCS Events.
    """

    _unittest_name = "vcsevents"
    _endpoint_being_tested = "vcs_events"


    def setUp(self):
        self._oauth_client = self._api.oauth_clients.create(\
            self._get_oauth_client_create_payload())["data"]
        self._oauth_client_id = self._oauth_client["id"]

    def tearDown(self):
        self._api.oauth_clients.destroy(self._oauth_client_id)

    def test_vcs_events(self):
        """
        Test the VCS Events API endpoints.
        """
        # NOTE: Currently, this feature is in beta, and only supports GitLab.
        # Since all of the testing is done with GitHub, this function is
        # provided, but you can see below that the test is not yet very good.

        # List the VCS events
        start_page = 0
        page_size = 50
        # TODO: test filter / include parameters GitHub is supported.
        vcs_events = self._api.vcs_events.list(page=start_page, page_size=page_size)["data"]
        self.assertEqual(len(vcs_events), 0)



"""
Module for testing the Terraform Cloud API Endpoint: Config Versions.
"""

import time

from .base import TestTFCBaseTestCase


class TestTFCConfigVersions(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Config Versions.
    """

    _unittest_name = "cnf-ver"
    _endpoint_being_tested = "config_versions"

    def setUp(self):
        self._ws = self._api.workspaces.create(self._get_ws_no_vcs_create_payload())
        self._ws_id = self._ws["data"]["id"]

        oauth_client = self._api.oauth_clients.create(self._get_oauth_client_create_payload())
        self._oauth_client_id = oauth_client["data"]["id"]
        oauth_token_id = oauth_client["data"]["relationships"]["oauth-tokens"]["data"][0]["id"]

        ws_payload = self._get_ws_with_vcs_create_payload(oauth_token_id)
        self._ws_w_vcs = self._api.workspaces.create(ws_payload)
        self._ws_w_vcs_id = self._ws_w_vcs["data"]["id"]

    def tearDown(self):
        self._api.workspaces.destroy(workspace_id=self._ws_id)
        self._api.workspaces.destroy(workspace_id=self._ws_w_vcs_id)
        self._api.oauth_clients.destroy(self._oauth_client_id)

    def test_config_versions(self):
        """
        Test the Config Versions API endpoints.
        """

        # two upload methods to test; use a different 1st positional argument for each test
        upload_tests = (
            (self._api.config_versions.upload, self._config_version_upload_tarball_path),
            (self._api.config_versions.upload_from_string, self._config_version_upload_string),
        )
        for upload_handle, source in upload_tests:

            with self.subTest():

                # Create a new config version
                created_config_version = self._api.config_versions.create(
                    self._ws_id, self._get_config_version_create_payload())["data"]
                cv_id = created_config_version["id"]

                # List all of the config versions for the workspace
                config_versions = self._api.config_versions.list(self._ws_id)["data"]

                # Confirm we found the newly created config version
                found_conf_ver = False
                for conf_ver in config_versions:
                    if cv_id == conf_ver["id"]:
                        found_conf_ver = True
                        break
                self.assertTrue(found_conf_ver)

                # Confirm the config version status is pending and the IDs match on show
                shown_config_version = self._api.config_versions.show(cv_id)["data"]
                self.assertEqual(shown_config_version["attributes"]["status"], "pending")
                self.assertEqual(shown_config_version["id"], cv_id)

                # Upload the .tf code and confirm it's been uploaded
                upload_url = created_config_version["attributes"]["upload-url"]
                upload_handle(source, upload_url)

                # Give 2 seconds for the upload to register
                time.sleep(2)
                config_versions = self._api.config_versions.list(self._ws_id)["data"]
                self.assertEqual(config_versions[0]["attributes"]["status"], "uploaded")

                all_config_versions = self._api.config_versions.list_all(self._ws_id)
                found_conf_ver = False
                for conf_ver in all_config_versions["data"]:
                    if cv_id == conf_ver["id"]:
                        found_conf_ver = True
                        break
                self.assertTrue(found_conf_ver)

    def test_config_versions_includes(self):
        """
        Test the related resources for the Config Versions API endpoints (requires VCS).
        """
        config_versions_raw = self._api.config_versions.list(self._ws_w_vcs_id, \
            include=["ingress-attributes"])
        self.assertIn("included", config_versions_raw)

        # Grab the config version ID from the list, and show it.
        cv_id = config_versions_raw["data"][0]["id"]
        shown_config_version_raw = self._api.config_versions.show(cv_id, \
            include=["ingress_attributes"])
        # Confirm we get the related resources.
        self.assertIn("included", shown_config_version_raw)

        # Show the commit information for the config version
        shown_commit_info = self._api.config_versions.show_config_version_commit_info(cv_id)["data"]
        self.assertIn("github", shown_commit_info["attributes"]["commit-url"])

        all_config_versions_raw = self._api.config_versions.list_all(self._ws_w_vcs_id, \
            include=["ingress-attributes"])
        self.assertIn("included", all_config_versions_raw)



"""
Module for testing the Terraform Cloud API Endpoint: Org Tags.
"""

from .base import TestTFCBaseTestCase


class TestTFCOrgTags(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Org Tags.
    """

    _unittest_name = "org-tag"
    _endpoint_being_tested = "org_tags"

    def setUp(self):
        self._ws_0 = self._api.workspaces.create(self._get_ws_no_vcs_create_payload())
        self._ws_0_id = self._ws_0["data"]["id"]
        self._ws_0_name = self._ws_0["data"]["attributes"]["name"]

        self._ws_1 = self._api.workspaces.create(self._get_ws_no_vcs_create_payload())
        self._ws_1_id = self._ws_1["data"]["id"]
        self._ws_1_name = self._ws_1["data"]["attributes"]["name"]

        self._ws_0_add_tags_payload = {
            "data": [
                {
                    "type": "tags",
                    "attributes": {
                        "name": "foo"
                    }
                },
                {
                    "type": "tags",
                    "attributes": {
                        "name": "bar"
                    }
                }
            ]
        }

        self._api.workspaces.add_tags(self._ws_0_id, self._ws_0_add_tags_payload)

    def tearDown(self):
        self._api.workspaces.destroy(workspace_name=self._ws_0_name)
        self._api.workspaces.destroy(workspace_name=self._ws_1_name)

    def test_org_tags(self):
        """
        Test the Org Tags API endpoints.
        """
        # Get the tags that were added in the setup, confirm both were created.
        org_tags = self._api.org_tags.list_tags()["data"]
        tag_0_id = org_tags[0]["id"]
        tag_1_id = org_tags[1]["id"]
        self.assertEqual(len(org_tags), len(self._ws_0_add_tags_payload["data"]))

        # Add one of the tags to workspace 1, confirm it has been attached.
        add_ws_to_tag_payload = {
            "data": [
                {
                    "type": "workspaces",
                    "id": self._ws_1_id
                }
            ]
        }
        self._api.org_tags.add_workspaces_to_tag(tag_0_id, add_ws_to_tag_payload)
        ws_1_tags = self._api.workspaces.list_tags(self._ws_1_id)["data"]
        self.assertEqual(tag_0_id, ws_1_tags[0]["id"])

        """
        # FIXME: this isn't working properly, the tag is still present on the workspace
        # Remove that tag from workspace 1, confirm it has been removed.
        remove_ws_from_tag_payload = {
            "data": [
                {
                    "type": "workspaces",
                    "id": self._ws_1_id
                }
            ]
        }
        self._api.org_tags.remove_workspaces_from_tag(tag_0_id, remove_ws_from_tag_payload)
        ws_1_tags = self._api.workspaces.list_tags(self._ws_1_id)["data"]
        print(ws_1_tags)
        self.assertEqual(len(ws_1_tags), 0)
        """

        # Delete both of the tags and confirm that workspace 0 no longer has the tags.
        delete_tags_payload = {
            "data": [
                {
                    "type": "tags",
                    "id": tag_0_id
                },
                {
                    "type": "tags",
                    "id": tag_1_id
                }
            ]
        }
        self._api.org_tags.delete_tags(delete_tags_payload)
        ws_1_tags = self._api.workspaces.list_tags(self._ws_0_id)["data"]
        self.assertEqual(len(ws_1_tags), 0)



# pylint: disable=too-many-instance-attributes

"""
Module for container class of all TFC endpoints and high level exceptions around
API access.
"""

from urllib.parse import urlparse
import json
import logging
import requests
import urllib3

from ._constants import \
    TFC_SAAS_URL, TFC_SAAS_HOSTNAME, HTTP_OK, TERRASNEK_LOG_LEVEL, TERRASNEK_VERSION
from .exceptions import TFCHTTPNotFound

from .account import TFCAccount
from .admin_module_sharing import TFCAdminModuleSharing
from .admin_orgs import TFCAdminOrgs
from .admin_runs import TFCAdminRuns
from .admin_settings import TFCAdminSettings
from .admin_terraform_versions import TFCAdminTerraformVersions
from .admin_users import TFCAdminUsers
from .admin_workspaces import TFCAdminWorkspaces
from .agents import TFCAgents
from .agent_tokens import TFCAgentTokens
from .applies import TFCApplies
from .audit_trails import TFCAuditTrails
from .config_versions import TFCConfigVersions
from .cost_estimates import TFCCostEstimates
from .feature_sets import TFCFeatureSets
from .ip_ranges import TFCIPRanges
from .invoices import TFCInvoices
from .oauth_clients import TFCOAuthClients
from .oauth_tokens import TFCOAuthTokens
from .orgs import TFCOrgs
from .org_memberships import TFCOrgMemberships
from .org_tags import TFCOrgTags
from .org_tokens import TFCOrgTokens
from .plans import TFCPlans
from .plan_exports import TFCPlanExports
from .policies import TFCPolicies
from .policy_checks import TFCPolicyChecks
from .policy_sets import TFCPolicySets
from .policy_set_params import TFCPolicySetParams
from .notification_configs import TFCNotificationConfigurations
from .registry_modules import TFCRegistryModules
from .run_tasks import TFCRunTasks
from .run_triggers import TFCRunTriggers
from .runs import TFCRuns
from .state_versions import TFCStateVersions
from .state_version_outputs import TFCStateVersionOutputs
from .ssh_keys import TFCSSHKeys
from .subscriptions import TFCSubscriptions
from .teams import TFCTeams
from .team_access import TFCTeamAccess
from .team_memberships import TFCTeamMemberships
from .team_tokens import TFCTeamTokens
from .users import TFCUsers
from .user_tokens import TFCUserTokens
from .vars import TFCVars
from .vcs_events import TFCVCSEvents
from .workspace_vars import TFCWorkspaceVars
from .workspaces import TFCWorkspaces

# Suppress insecure TLS warnings
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class InvalidTFCTokenException(Exception):
    """Cannot instantiate TFC API class without a valid TFC_TOKEN."""


class TFC():
    """
    Super class for access to all TFC Endpoints.
    """

    # This dict specifies which class should be used for each attribute of this API class,
    # which simplifies the initialization of each endpoint since they share the same initialization
    # values.

    _class_for_attr_dict = {
        "org-not-required": {
            "admin_orgs": TFCAdminOrgs,
            "admin_module_sharing": TFCAdminModuleSharing,
            "admin_runs": TFCAdminRuns,
            "admin_settings": TFCAdminSettings,
            "admin_terraform_versions": TFCAdminTerraformVersions,
            "admin_users": TFCAdminUsers,
            "admin_workspaces": TFCAdminWorkspaces,
            "audit_trails": TFCAuditTrails,
            "ip_ranges": TFCIPRanges,
            "orgs": TFCOrgs
        },
        "org-required": {
            "account": TFCAccount,
            "applies": TFCApplies,
            "agents": TFCAgents,
            "agent_tokens": TFCAgentTokens,
            "config_versions": TFCConfigVersions,
            "cost_estimates": TFCCostEstimates,
            "feature_sets": TFCFeatureSets,
            "invoices": TFCInvoices,
            "oauth_clients": TFCOAuthClients,
            "oauth_tokens": TFCOAuthTokens,
            "org_memberships": TFCOrgMemberships,
            "org_tags": TFCOrgTags,
            "org_tokens": TFCOrgTokens,
            "plans": TFCPlans,
            "plan_exports": TFCPlanExports,
            "policies": TFCPolicies,
            "policy_checks": TFCPolicyChecks,
            "policy_sets": TFCPolicySets,
            "policy_set_params": TFCPolicySetParams,
            "notification_configs": TFCNotificationConfigurations,
            "registry_modules": TFCRegistryModules,
            "run_tasks": TFCRunTasks,
            "run_triggers": TFCRunTriggers,
            "runs": TFCRuns,
            "ssh_keys": TFCSSHKeys,
            "state_versions": TFCStateVersions,
            "state_version_outputs": TFCStateVersionOutputs,
            "subscriptions": TFCSubscriptions,
            "teams": TFCTeams,
            "team_access": TFCTeamAccess,
            "team_memberships": TFCTeamMemberships,
            "team_tokens": TFCTeamTokens,
            "users": TFCUsers,
            "user_tokens": TFCUserTokens,
            "vars": TFCVars,
            "vcs_events": TFCVCSEvents,
            "workspace_vars": TFCWorkspaceVars,
            "workspaces": TFCWorkspaces
        }
    }

    def __init__(self, api_token, url=TFC_SAAS_URL, verify=True, log_level=TERRASNEK_LOG_LEVEL):
        if api_token is None:
            raise InvalidTFCTokenException

        self._logger = logging.getLogger(self.__class__.__name__)
        self._log_level = log_level
        self._logger.setLevel(self._log_level)

        self._logger.debug("Initializing the TFC API class...")

        self._instance_url = url

        parsed_url = urlparse(url)
        self._hostname = parsed_url.netloc

        self._token = api_token
        self._current_org = None
        self._verify = verify

        self.__version__ = TERRASNEK_VERSION
        self.version = TERRASNEK_VERSION

        self.admin_module_sharing: TFCAdminModuleSharing = None
        self.admin_orgs: TFCAdminOrgs = None
        self.admin_runs: TFCAdminRuns = None
        self.admin_settings: TFCAdminSettings = None
        self.admin_terraform_versions: TFCAdminTerraformVersions = None
        self.admin_users: TFCAdminUsers = None
        self.admin_workspaces: TFCAdminWorkspaces = None
        self.audit_trails: TFCAuditTrails = None
        self.invoices: TFCInvoices = None
        self.ip_ranges: TFCIPRanges = None
        self.feature_sets: TFCFeatureSets = None
        self.orgs: TFCOrgs = None
        self.account: TFCAccount = None
        self.applies: TFCApplies = None
        self.agents: TFCAgents = None
        self.agent_tokens: TFCAgentTokens = None
        self.config_versions: TFCConfigVersions = None
        self.cost_estimates: TFCCostEstimates = None
        self.oauth_clients: TFCOAuthClients = None
        self.oauth_tokens: TFCOAuthTokens = None
        self.org_memberships: TFCOrgMemberships = None
        self.org_tags: TFCOrgTags = None
        self.org_tokens: TFCOrgTokens = None
        self.plans: TFCPlans = None
        self.plan_exports: TFCPlanExports = None
        self.policies: TFCPolicies = None
        self.policy_checks: TFCPolicyChecks = None
        self.policy_sets: TFCPolicySets = None
        self.policy_set_params: TFCPolicySetParams = None
        self.notification_configs: TFCNotificationConfigurations = None
        self.registry_modules: TFCRegistryModules = None
        self.run_tasks: TFCRunTasks = None
        self.run_triggers: TFCRunTriggers = None
        self.runs: TFCRuns = None
        self.state_versions: TFCStateVersions = None
        self.state_version_outputs: TFCStateVersionOutputs = None
        self.ssh_keys: TFCSSHKeys = None
        self.teams: TFCTeams = None
        self.team_access: TFCTeamAccess = None
        self.team_memberships: TFCTeamMemberships = None
        self.team_tokens: TFCTeamTokens = None
        self.users: TFCUsers = None
        self.user_tokens: TFCUserTokens = None
        self.vars: TFCVars = None
        self.vcs_events: TFCVCSEvents = None
        self.workspace_vars: TFCWorkspaceVars = None
        self.workspaces: TFCWorkspaces = None

        self._token = None
        self._headers = None

        self._logger.debug("Retrieving TFC API well known paths..")
        self._well_known_paths = self.well_known_paths()
        self._logger.debug("TFC API well known paths retrieved.")

        self.set_token(api_token)
        self._initialize_endpoints()

    def _get(self, url):
        """
        Simplified HTTP GET function for usage only with this API module.
        """
        results = None
        req = requests.get(url, headers=self._headers, verify=self._verify)

        if req.status_code == HTTP_OK:
            results = json.loads(req.content)
            self._logger.debug(f"GET to {url} successful")
        else:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)

        return results

    def _initialize_endpoints(self):
        self._logger.debug("Initializing endpoints that don't require an org to be set...")
        # Loop through all the endpoints that don't require an org and initialize them
        for ep_name in self._class_for_attr_dict["org-not-required"]:
            class_for_attr = self._class_for_attr_dict["org-not-required"][ep_name]
            initialized_endpoint_class = class_for_attr(
                self._instance_url,
                None,
                self._headers,
                self._well_known_paths,
                self._verify,
                self._log_level)
            setattr(self, ep_name, initialized_endpoint_class)

        self._logger.debug("Initialized endpoints that don't require an org to be set.")
        # Loop through all the endpoints that do require an org and initialize them
        for ep_name in self._class_for_attr_dict["org-required"]:
            setattr(self, ep_name, None)

        self._logger.debug("TFC API class initialized.")

    def get_url(self):
        """
        Allows for the user to retrieve the URL being hit from the API object.
        """
        return self._instance_url

    def get_hostname(self):
        """
        Allows for the user to retrieve the hostname being hit from the API object.
        """
        return self._hostname

    def set_org(self, org_name):
        """
        Sets the organization to use for org specific endpoints.
        This method must be called for any non-admin endpoint to work.
        """

        self._logger.debug("Initializing endpoints that do require an org to be set...")

        # Update the current org attribute
        self._current_org = org_name

        for ep_name in self._class_for_attr_dict["org-required"]:
            class_for_attr = self._class_for_attr_dict["org-required"][ep_name]
            initialized_endpoint_class = class_for_attr(
                self._instance_url,
                self._current_org,
                self._headers,
                self._well_known_paths,
                self._verify,
                self._log_level)

            setattr(self, ep_name, initialized_endpoint_class)

        self._logger.debug("Initialized endpoints that do require an org to be set.")

    def get_org(self):
        """
        Allows for the user to retrieve the current org from the API object.
        """
        return self._current_org

    def get_entitlements(self):
        """
        Allows for the user to retrieve the entitlements to the API for the current org.
        """
        entitlements = None

        if self.is_terraform_cloud():
            try:
                entitlements = self.orgs.entitlements(self._current_org)["data"]["attributes"]
            except TFCHTTPNotFound:
                self._logger.debug("Entitlements API endpoint not found. No entitlements recorded.")
        else:
            self._logger.debug("Not Terraform Cloud, so entitlements API is not supported.")

        return entitlements

    def set_token(self, token):
        """
        Allows for the user to change the token they are using on the fly if
        they need to change tokens.
        """
        self._token = token
        self._headers = {
            "Authorization": f"Bearer {token}",
            "User-Agent": f"terrasnek-{self.version}",
            "Content-Type": "application/vnd.api+json"
        }
        self._initialize_endpoints()

    def get_token(self):
        """
        Allows for the user to retrieve the token from the API object.
        """
        return self._token

    def is_terraform_cloud(self):
        """
        Returns true if this API instance is configured for Terraform Cloud.
        """
        return TFC_SAAS_HOSTNAME in self._instance_url


    def well_known_paths(self):
        """
        Retrieve all the well known paths from the Terraform Cloud installation.
        """
        url = f"{self._instance_url}/.well-known/terraform.json"
        return self._get(url)



"""
Module for testing the Terraform Cloud API Endpoint: Run Tasks.
"""

from .base import TestTFCBaseTestCase


class TestTFCRunTasks(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Run Tasks.
    """

    _unittest_name = "run-task"
    _endpoint_being_tested = "run_tasks"

    def setUp(self):
        self._ws = self._api.workspaces.create(self._get_ws_no_vcs_create_payload())
        self._ws_id = self._ws["data"]["id"]
        self._ws_name = self._ws["data"]["attributes"]["name"]

    def tearDown(self):
        self._api.workspaces.destroy(workspace_name=self._ws_name)

    def test_run_tasks(self):
        """
        Test the Run Tasks API endpoints.
        """
        # List the event hooks, check that there are none
        hooks = self._api.run_tasks.list_event_hooks(self._ws_id)["data"]
        self.assertEqual(len(hooks), 0)

        # Create an event hook
        hook_name = "example"
        create_hook_payload = {
            "data": {
                "type": "event-hooks",
                "attributes": {
                    "name": hook_name,
                    "url": "http://example.com",
                    "hmac_key": "secret",
                    "category": "task"
                }
            }
        }
        created_hook = self._api.run_tasks.create_event_hook(create_hook_payload)["data"]
        created_hook_id = created_hook["id"]
        self.assertEqual(created_hook["attributes"]["name"], hook_name)

        # Confirm an event hook was created
        shown_hook = self._api.run_tasks.show_event_hook(created_hook_id)["data"]
        self.assertEqual(shown_hook["attributes"]["name"], hook_name)

        # List the event hooks, check that we get the one we created
        listed_hooks = self._api.run_tasks.list_event_hooks()["data"]
        self.assertEqual(len(listed_hooks), 1)

        # List all the event hooks, check that we get the one we created
        all_hooks = self._api.run_tasks.list_all_event_hooks()["data"]
        self.assertEqual(len(all_hooks), 1)

        updated_hook_name = "new-example"
        update_hook_payload = {
            "data": {
                "type": "event-hooks",
                "attributes": {
                    "name": updated_hook_name
                }
            }
        }
        updated_hook = self._api.run_tasks.update_event_hook(created_hook_id, update_hook_payload)["data"]
        self.assertEqual(updated_hook["attributes"]["name"], updated_hook_name)

        # Attach the event hook as a task to a workspace
        attach_hook_payload = {
            "data": {
                "type": "tasks",
                "attributes": {
                    "enforcement-level": "advisory"
                },
                "relationships": {
                    "event-hook": {
                        "data": {
                            "id": created_hook_id,
                            "type": "event-hooks"
                        }
                    }
                }
            }
        }
        attached_task = \
            self._api.run_tasks.attach_event_hook_as_task(self._ws_id, attach_hook_payload)["data"]
        attached_task_id = attached_task["id"]
        attached_task_hook_id = attached_task["relationships"]["event-hook"]["data"]["id"]
        self.assertEqual(created_hook_id, attached_task_hook_id)

        # List the tasks, confirm we have one
        listed_tasks = self._api.run_tasks.list(self._ws_id)["data"]
        self.assertEqual(len(listed_tasks), 1)

        # Confirm that the hook has been attached by showing the task and comparing IDs
        shown_task = self._api.run_tasks.show(attached_task_id)["data"]
        self.assertEqual(shown_task["id"], attached_task_id)

        # Update the task
        updated_enforcement_level = "mandatory"
        update_task_payload = {
            "data": {
                "type": "tasks",
                "attributes": {
                    "enforcement-level": updated_enforcement_level
                }
            }
        }
        updated_task = self._api.run_tasks.update(attached_task_id, update_task_payload)["data"]
        self.assertEqual(updated_task["attributes"]["enforcement-level"], updated_enforcement_level)

        # Destroy the task, confirm it's gone
        self._api.run_tasks.destroy(attached_task_id)
        listed_tasks = self._api.run_tasks.list(self._ws_id)["data"]
        self.assertEqual(len(listed_tasks), 0)

        # Destroy the hook, confirm it's gone
        self._api.run_tasks.destroy_event_hook(created_hook_id)
        listed_hooks = self._api.run_tasks.list_event_hooks()["data"]
        self.assertEqual(len(listed_hooks), 0)



"""
Module containing class for common endpoint implementations across all TFC Endpoints.
"""

from abc import ABC, abstractmethod

import json
import logging
import requests

from .exceptions import \
    TFCHTTPBadRequest, TFCHTTPUnauthorized, TFCHTTPForbidden, TFCHTTPNotFound, \
        TFCHTTPConflict, TFCHTTPPreconditionFailed, TFCHTTPUnprocessableEntity, \
            TFCHTTPInternalServerError, TFCHTTPUnclassified, TFCHTTPAPIRequestRateLimit

from ._constants import \
    HTTP_OK, HTTP_CREATED, HTTP_ACCEPTED, HTTP_NO_CONTENT, HTTP_BAD_REQUEST, HTTP_UNAUTHORIZED, \
        HTTP_FORBIDDEN, HTTP_NOT_FOUND, HTTP_CONFLICT, HTTP_PRECONDITION_FAILED, \
            HTTP_UNPROCESSABLE_ENTITY, HTTP_API_REQUEST_RATE_LIMIT_REACHED, \
                HTTP_INTERNAL_SERVER_ERROR, MAX_PAGE_SIZE

class TFCEndpoint(ABC):
    """
    Base class providing common CRUD operation implementations across all TFC Endpoints.
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._logger.setLevel(log_level)

        # Remove the slack at the end if someone adds it.
        self._instance_url = \
            instance_url if instance_url[-1] != "/" else instance_url[:-1]
        self._api_v2_base_url = f"{self._instance_url}{well_known_paths['tfe.v2'][:-1]}"
        self._meta_base_url = f"{self._instance_url}/api/meta"
        self._modules_v1_base_url = f"{self._instance_url}{well_known_paths['modules.v1'][:-1]}"
        self._headers = headers
        self._org_name = org_name
        self._verify = verify

    @abstractmethod
    def required_entitlements(self):
        """
        Terraform Cloud Entitlements required for endpoint to work.
        """
        return []

    @abstractmethod
    def terraform_cloud_only(self):
        """
        Return ``True`` if this endpoint is only for Terraform Cloud, else ``False``.
        """
        return False

    @abstractmethod
    def terraform_enterprise_only(self):
        """
        Return ``True`` if this endpoint is only for Terraform Enterprise, else ``False``.
        """
        return False

    def _delete(self, url, data=None):
        results = None
        req = requests.delete(\
            url, data=json.dumps(data), headers=self._headers, verify=self._verify)

        if req.status_code == HTTP_NO_CONTENT:
            pass
        elif req.status_code == HTTP_NOT_FOUND:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPNotFound(err)
        elif req.status_code == HTTP_FORBIDDEN:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPForbidden(err)
        elif req.status_code == HTTP_API_REQUEST_RATE_LIMIT_REACHED:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPAPIRequestRateLimit(err)
        else:
            try:
                err = json.loads(req.content.decode("utf-8"))
            except json.decoder.JSONDecodeError:
                err = req.content
            self._logger.debug(err)
            raise TFCHTTPUnclassified(err)

        return results

    def _get(self, url, return_raw=None, allow_redirects=None, query=None, filters=None, \
        page=None, page_size=None, search=None, include=None, sort=None, \
        offset=None, limit=None, provider=None, namespace=None, verified=None, \
        since=None):

        results = None

        q_options = []

        if query is not None:
            q_options.append(f"q={query}")

        if filters is not None:
            for fil in filters:
                filter_string = "filter"
                for k in fil["keys"]:
                    filter_string += f"[{k}]"
                filter_string += f"={fil['value']}"
                q_options.append(filter_string)

        if page is not None:
            q_options.append(f"page[number]={page}")

        if page_size is not None:
            q_options.append(f"page[size]={page_size}")

        if search is not None:
            q_options.append(f"page[name]={page_size}")

        if include is not None:
            joined_include = ",".join(include)
            q_options.append(f"include={joined_include}")

        if sort is not None:
            q_options.append(f"sort={sort}")

        if search is not None:
            q_options.append(f"search[name]={search}")

        if since is not None:
            q_options.append(f"since={since}")

        # V1 Modules API options
        # TODO: do we still need to support this?
        if offset is not None:
            q_options.append(f"offset={offset}")

        if limit is not None:
            q_options.append(f"limit={limit}")

        if provider is not None:
            q_options.append(f"provider={provider}")

        if namespace is not None:
            q_options.append(f"namespace={namespace}")

        if verified is not None:
            q_options.append(f"verified={verified}")

        if q_options:
            url += "?" + "&".join(q_options)

        self._logger.debug(f"Trying HTTP GET to URL: {url} ...")
        req = requests.get(\
            url, headers=self._headers, verify=self._verify, allow_redirects=allow_redirects)

        if req.status_code == HTTP_OK and not return_raw:
            results = json.loads(req.content)
            self._logger.debug(f"GET to {url} successful")
        elif req.status_code == HTTP_OK and return_raw:
            results = req.content
        elif req.status_code == HTTP_NO_CONTENT:
            results = req.headers
        elif req.history:
            # NOTE: If we got redirected, run the get on the new URL, and fix the
            # URL to match the private module registry URL schema.
            url = req.url.replace("/v1/modules/", "/api/registry/v1/modules/")
            results = {"redirect-url": url}
        elif req.status_code == HTTP_UNAUTHORIZED:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPUnauthorized(err)
        elif req.status_code == HTTP_NOT_FOUND:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPNotFound(err)
        elif req.status_code == HTTP_API_REQUEST_RATE_LIMIT_REACHED:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPAPIRequestRateLimit(err)
        else:
            try:
                err = json.loads(req.content.decode("utf-8"))
            except json.decoder.JSONDecodeError:
                err = req.content
            self._logger.debug(err)
            raise TFCHTTPUnclassified(err)

        return results

    def _patch(self, url, data=None):
        results = None

        self._logger.debug(f"Trying HTTP PATCH to URL: {url} ...")
        req = requests.patch(url, data=json.dumps(data), headers=self._headers, verify=self._verify)

        if req.status_code == HTTP_OK:
            results = json.loads(req.content)
        elif req.status_code == HTTP_BAD_REQUEST:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPBadRequest(err)
        elif req.status_code == HTTP_UNAUTHORIZED:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPUnauthorized(err)
        elif req.status_code == HTTP_UNPROCESSABLE_ENTITY:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPUnprocessableEntity(err)
        elif req.status_code == HTTP_API_REQUEST_RATE_LIMIT_REACHED:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPAPIRequestRateLimit(err)
        elif req.status_code == HTTP_NO_CONTENT:
            pass
        else:
            try:
                err = json.loads(req.content.decode("utf-8"))
            except json.decoder.JSONDecodeError:
                err = req.content
            self._logger.debug(err)
            raise TFCHTTPUnclassified(err)

        return results

    def _post(self, url, data=None):
        results = None

        self._logger.debug(f"Trying HTTP POST to URL: {url} ...")
        req = requests.post(url, data=json.dumps(data), headers=self._headers, verify=self._verify)

        if req.status_code in [HTTP_OK, HTTP_CREATED]:
            results = json.loads(req.content)
            self._logger.debug(f"POST to {url} successful")
        elif req.status_code in [HTTP_ACCEPTED, HTTP_NO_CONTENT]:
            self._logger.debug(f"POST to {url} successful")
        elif req.status_code == HTTP_BAD_REQUEST:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPBadRequest(err)
        elif req.status_code == HTTP_NOT_FOUND:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPNotFound(err)
        elif req.status_code == HTTP_CONFLICT:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPConflict(err)
        elif req.status_code == HTTP_PRECONDITION_FAILED:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPPreconditionFailed(err)
        elif req.status_code == HTTP_UNPROCESSABLE_ENTITY:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPUnprocessableEntity(err)
        elif req.status_code == HTTP_API_REQUEST_RATE_LIMIT_REACHED:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPAPIRequestRateLimit(err)
        elif req.status_code == HTTP_INTERNAL_SERVER_ERROR:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPInternalServerError(err)
        else:
            try:
                err = json.loads(req.content.decode("utf-8"))
            except json.decoder.JSONDecodeError:
                err = req.content
            self._logger.debug(err)
            raise TFCHTTPUnclassified(err)

        return results

    def _put(self, url, octet=False, data=None):
        results = None

        headers = dict.copy(self._headers)
        if octet is True:
            headers["Content-Type"] = "application/octet-stream"
            data = bytes(data, "utf-8")

        self._logger.debug(f"Trying HTTP PUT to URL: {url} ...")
        req = requests.put(url, data=data, headers=headers, verify=self._verify)

        if req.status_code == HTTP_OK:
            if octet:
                results = json.loads(req.content)
            self._logger.debug(f"PUT to {url} successful")
        elif req.status_code == HTTP_API_REQUEST_RATE_LIMIT_REACHED:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPAPIRequestRateLimit(err)
        else:
            try:
                err = json.loads(req.content.decode("utf-8"))
            except json.decoder.JSONDecodeError:
                err = req.content
            self._logger.debug(err)
            raise TFCHTTPUnclassified(err)

        return results

    def _create(self, url, payload=None):
        """
        Implementation the common create resource pattern for the TFC API.
        """
        if payload is None:
            payload = {}

        return self._post(url, data=payload)

    def _destroy(self, url):
        """
        Implementation of the common destroy resource pattern for the TFC API.
        """
        req = requests.delete(url, headers=self._headers, verify=self._verify)

        valid_status_codes = [HTTP_OK, HTTP_NO_CONTENT]
        if req.status_code in valid_status_codes:
            self._logger.debug(f"Terraform Cloud resource at URL [{url}] destroyed.")
        elif req.status_code == HTTP_API_REQUEST_RATE_LIMIT_REACHED:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPAPIRequestRateLimit(err)
        else:
            try:
                err = json.loads(req.content.decode("utf-8"))
            except json.decoder.JSONDecodeError:
                err = req.content
            self._logger.debug(err)
            raise TFCHTTPUnclassified(err)

    def _list(self, url, query=None, filters=None, \
        page=None, page_size=None, search=None, include=None, sort=None, \
        offset=None, limit=None, provider=None, namespace=None, verified=None, \
        since=None):
        """
        Implementation of the common list resources pattern for the TFC API.
        """
        return self._get(url, query=query, filters=filters, page=page, \
            page_size=page_size, search=search, include=include, sort=sort, \
                offset=offset, limit=limit, provider=provider, namespace=namespace, \
                    verified=verified, since=since)

    def _list_all(self, url, include=None, search=None, filters=None, query=None):
        """
        This function does not correlate to an endpoint in the TFC API Docs specifically,
        but rather is a helper function to wrap the `list` endpoint, which enumerates out
        every page so users do not have to implement the paging logic every time they just
        want to list every workspace in an organization.

        Returns an object with two arrays of objects.
        """
        current_page_number = 1
        list_resp = \
            self._list(url, page=current_page_number, page_size=MAX_PAGE_SIZE, include=include, \
                search=search, filters=filters, query=query)

        if "meta" in list_resp:
            total_pages = list_resp["meta"]["pagination"]["total-pages"]
        elif "pagination" in list_resp:
            total_pages = list_resp["pagination"]["total_pages"]

        included = []
        data = []
        while current_page_number <= total_pages:
            list_resp = \
                self._list(url, page=current_page_number, page_size=MAX_PAGE_SIZE, \
                    include=include, search=search, filters=filters, query=query)
            data += list_resp["data"]

            if "included" in list_resp:
                included += list_resp["included"]

            current_page_number += 1

        return {
            "data": data,
            "included": included
        }

    def _show(self, url, include=None):
        """
        Implementation of the common show resource pattern for the TFC API.
        """
        return self._get(url, include=include)

    def _update(self, url, payload):
        """
        Implementation of the common update resource pattern for the TFC API.
        """
        return self._patch(url, data=payload)

    def _download(self, url, target_path, header_with_url=None, allow_redirects=False):
        """
        Implementation of a common download pattern from the TFC API.
        """
        results = None
        if header_with_url is None:
            results = self._get(url, return_raw=True, allow_redirects=allow_redirects)
        else:
            response = self._get(url, allow_redirects=allow_redirects)
            if "redirect-url" in response:
                response = self._get(response["redirect-url"])
            download_url = response[header_with_url]
            results = self._get(download_url, return_raw=True)

        with open(target_path, 'wb') as target_file:
            target_file.write(results)

    def get_current_org(self):
        """
        Return the active org for this endpoint.
        """
        return self._org_name
